from math import tanh
from typing import TYPE_CHECKING, List


if TYPE_CHECKING:
    from .models import Player, Team


def _normalize(x: float) -> float:
    # Force a number to be between 0 and 1
    return 1 - tanh(x)


class CountScorer:
    def __init__(self, players: List["Player"], teams: List["Team"]):
        self.league_size = len(players)
        self.count = 0
        for t in teams:
            self.count += sum([self._count_player(p, t) for p in t.players])

    def _count_player(self, player: "Player", team: "Team") -> bool:
        raise NotImplementedError

    def update_score_addition(self, player: "Player", team: "Team"):
        if self._count_player(player, team):
            self.count += 1

    def update_score_removal(self, player: "Player", team: "Team"):
        if self._count_player(player, team):
            self.count -= 1

    def get_score(self) -> float:
        return self.count / self.league_size


def _calculate_rmse(values, ideal_value):
    """Calculate the root mean squared error from a list of values and ideal value"""
    squared_errors = [(v - ideal_value) ** 2 for v in values]
    return (sum(squared_errors) / len(squared_errors)) ** 0.5


def _update_rolling_mean(old_mean, old_size, value, add_value):
    """Calculate the updated rolling mean for added or removed values."""
    if add_value:
        return (old_size * old_mean + value) / (old_size + 1)
    else:
        if old_size == 1:
            # If there aren't any more points, return 0 to avoid a divide-by-zero error
            return 0
        return (old_size * old_mean - value) / (old_size - 1)


class CountParityScorer:
    def __init__(self, players: List["Player"], teams: List["Team"]):
        self.total_player_count = 0
        self.team_counts = {team.name: 0 for team in teams}
        for t in teams:
            team_count = sum([self._count_player(p) for p in t.players])
            self.total_player_count += team_count
            self.team_counts[t.name] += team_count

    def _count_player(self, player: "Player") -> bool:
        raise NotImplementedError

    def update_score_addition(self, player: "Player", team: "Team"):
        if self._count_player(player):
            self.total_player_count += 1
            self.team_counts[team.name] += 1

    def update_score_removal(self, player: "Player", team: "Team"):
        if self._count_player(player):
            self.total_player_count -= 1
            self.team_counts[team.name] -= 1

    def get_score(self) -> float:
        ideal_team_count = self.total_player_count / len(self.team_counts)
        rmse = _calculate_rmse(self.team_counts.values(), ideal_team_count)
        return _normalize(rmse)


class MeanParityScorer:
    def __init__(self, players: List["Player"], teams: List["Team"]):
        self.total_value = 0
        self.team_means = {team.name: 0.0 for team in teams}
        for t in teams:
            team_mean = sum([self._get_value(p) for p in t.players])
            self.total_value += team_mean
            self.team_means[t.name] += team_mean
        self.total_players = sum([len(t.players) for t in teams])

    def _get_value(self, player: "Player") -> bool:
        raise NotImplementedError

    def update_score_addition(self, player: "Player", team: "Team"):
        self.total_value += self._get_value(player)
        self.total_players += 1
        self.team_means[team.name] = _update_rolling_mean(
            self.team_means[team.name],
            len(team.players),
            self._get_value(player),
            add_value=True,
        )

    def update_score_removal(self, player: "Player", team: "Team"):
        self.total_value -= self._get_value(player)
        self.total_players -= 1
        self.team_means[team.name] = _update_rolling_mean(
            self.team_means[team.name],
            len(team.players),
            self._get_value(player),
            add_value=False,
        )

    def get_score(self) -> float:
        ideal_value = self.total_value / self.total_players
        rmse = _calculate_rmse(self.team_means.values(), ideal_value)
        return _normalize(rmse)



import os
from typing import TYPE_CHECKING

from .utils import DAY_MAP, FIELD_MAP


if TYPE_CHECKING:
    from .models import Player, Team


def validate_player_strs(player: "Player"):
    for key in ["first_name", "last_name"]:
        value = player.__getattribute__(key)
        if not isinstance(value, str):
            raise ValueError(
                f"Failed to create player {player.first_name} {player.last_name}. Expected a "
                f"string but found {key}={value} ({type(value)}) instead. Please correct this in "
                "the input csv and retry"
            )
        if len(value) == 0:
            raise ValueError(
                f"Failed to create player {player.first_name} {player.last_name}. {key} has length "
                "0, but shouldn't be empty. Please correct this in the input csv and retry."
            )


def validate_player_ints(player: "Player"):
    for key, minimum, maximum in [
        ("grade", -1, 8),
        ("skill", 1, 10),
        ("goalie_skill", 1, 10),
    ]:
        value = player.__getattribute__(key)
        if not isinstance(value, int) and not isinstance(value, float):
            raise ValueError(
                f"Failed to create player {player.first_name} {player.last_name}. Expected a "
                f"number but found {key}={value} ({type(value)}) instead. Please correct this in "
                "the input csv and retry"
            )
        if value < minimum or value > maximum:
            raise ValueError(
                f"Failed to create player {player.first_name} {player.last_name}. {key} ({value}) "
                f"is unexpectedly < {minimum} or > {maximum}. Please correct this in the input csv "
                "and retry."
            )


def validate_player_days(player: "Player"):
    valid_days = "".join(list(DAY_MAP.values()))
    for day in player.unavailable_days + player.preferred_days:
        if day not in valid_days:
            raise ValueError(
                f"Failed to create player {player.first_name} {player.last_name} due to invalid "
                f"day. {day}  is not a valid day ({valid_days}). Please correct this in the input "
                "csv and retry."
            )

    for day in player.unavailable_days:
        if day in player.preferred_days:
            raise ValueError(
                f"Failed to create player {player.first_name} {player.last_name} due to invalid "
                f"day preferences: {day} was marked as both unavailable ({player.unavailable_days})"
                f" and preferred ({player.preferred_days}). Please correct this in the csv and "
                "retry."
            )


def validate_player_bools(player: "Player"):
    for key in ["lock", "emailed_parents"]:
        value = player.__getattribute__(key)
        if not isinstance(value, bool):
            raise ValueError(
                f"Failed to create player {player.first_name} {player.last_name}. Expected a "
                f"bool but found {key}={value} ({type(value)}) instead. Please correct this in "
                "the input csv and retry"
            )
    if player.emailed_parents and not player.lock:
        raise ValueError(
            f"Failed to create player {player.first_name} {player.last_name}. emailed_parents was "
            "set to True while lock was set to False. If parents have already been emailed, the "
            "player should be locked to their teams. Please correct this in the input csv and "
            "retry"
        )


def validate_player_locations(player: "Player"):
    valid_locations = set()
    for region, fields in FIELD_MAP.items():
        valid_locations.update(fields)

    # TODO: figure out how to strip strings to make the parsing more forgiving
    disallowed_locations = player.disallowed_locations.split(", ")
    preferred_locations = player.preferred_locations.split(", ")
    backup_locations = player.backup_locations.split(", ")
    for loc in disallowed_locations + preferred_locations + backup_locations:
        if len(loc) > 0 and loc not in valid_locations:
            raise ValueError(
                f"Failed to create player {player.first_name} {player.last_name} due to invalid "
                f"location. {loc} is not a valid location ({valid_locations}). Please correct in "
                "the input csv and retry."
            )

    for loc in disallowed_locations:
        if len(loc) > 0 and loc in player.preferred_locations:
            raise ValueError(
                f"Failed to create player {player.first_name} {player.last_name} ({player.id}) due "
                f"to invalid location preferences: {loc} was marked as both disallowed "
                f"({player.disallowed_locations}) and preferred ({player.preferred_locations}). "
                "Please correct this in the csv and retry."
            )


def validate_team_name(team: "Team"):
    if not isinstance(team.name, str):
        raise ValueError(
            f"Failed to create team {team}. Name should be a string, but was instead a "
            f"{type(team.name)}. Please correct the input csv and retry."
        )

    if len(team.name) == 0:
        raise ValueError(
            f"Failed to create team {team}. Name should not have length == 0. Please correct the "
            "input csv and retry."
        )


def validate_team_day(team: "Team"):
    valid_days = "".join(list(DAY_MAP.values()))
    if team.practice_day not in valid_days:
        raise ValueError(
            f"Failed to create team {team.name}. Practice day {team.practice_day} is not a valid "
            f"practice day ({valid_days}). Please orrect the input csv and retry."
        )


def validate_team_location(team: "Team"):
    valid_locations = set()
    for region, fields in FIELD_MAP.items():
        valid_locations.update(fields)

    if team.location not in valid_locations:
        raise ValueError(
            f"Failed to create team {team.name}. Location {team.location} is not in the valid "
            f"location list: {valid_locations}. Please check spelling, correct the input csv and "
            "retry"
        )


def validate_file(filename: str):
    if not os.path.exists(filename):
        raise FileNotFoundError(
            f"Could not find file: {filename}. Please check that you spelled the file name "
            "correctly and are running the command from the correct directory."
        )


def request_validation(message: str):
    print(message)
    _ = input("Press Enter to continue, or ctrl+c to quit:")



from dataclasses import asdict, dataclass, field
from typing import List, Optional, Set
from cyslf.utils import handle_error
import pandas as pd

from .constraints import breaks_practice_constraint
from .scorers import CompositeScorer
from .utils import (
    BOTTOM_TIER_SKILLS,
    FIRST_ROUND_SKILL,
    GOALIE_THRESHOLD,
    MID_TIER_SKILLS,
    TOP_TIER_SKILLS,
)
from .validation import (
    validate_player_bools,
    validate_player_days,
    validate_player_ints,
    validate_player_locations,
    validate_player_strs,
    validate_team_day,
    validate_team_location,
    validate_team_name,
)


@dataclass(frozen=True)
class Player:
    id: int
    first_name: str
    last_name: str
    grade: int
    skill: int
    coach_skill: int
    parent_skill: int
    goalie_skill: int
    unavailable_days: str
    preferred_days: str
    disallowed_locations: str
    preferred_locations: str
    backup_locations: str
    teammate_requests: str
    lock: bool
    emailed_parents: bool
    school: str
    comment: str

    @classmethod
    def from_raw_dict(cls, raw_dict):
        """Preprocess a raw csv row in dictionary form to create a Player."""
        # Convert nans to empty strings
        str_keys = [
            "unavailable_days",
            "preferred_days",
            "disallowed_locations",
            "preferred_locations",
            "backup_locations",
            "teammate_requests",
        ]
        for key in str_keys:
            if pd.isnull(raw_dict[key]):
                raw_dict[key] = ""

        raw_dict["skill"] = raw_dict["coach_skill"]
        if pd.isnull(raw_dict["skill"]):
            raw_dict["skill"] = raw_dict["parent_skill"]

        return cls(**raw_dict)

    def to_raw_dict(self) -> dict:
        """Create a raw dict that can be saved to a csv."""
        raw_dict = asdict(self)
        del raw_dict["skill"]
        return raw_dict

    def __post_init__(self):
        """Validate player data."""
        validate_player_strs(self)
        validate_player_ints(self)
        validate_player_bools(self)
        validate_player_days(self)
        validate_player_locations(self)



@dataclass
class Team:
    name: str
    practice_day: str
    location: str
    players: Set[Player] = field(default_factory=set)
    # TODO: practice time

    def get_skill(self) -> float:
        if len(self.players) == 0:
            return 0
        return sum([player.skill for player in self.players]) / len(self.players)

    def get_grade(self) -> float:
        if len(self.players) == 0:
            return 0
        return sum([player.grade for player in self.players]) / len(self.players)

    def get_first_round(self) -> int:
        return sum([p.skill == FIRST_ROUND_SKILL for p in self.players])

    def get_top_tier(self) -> int:
        return sum([p.skill in TOP_TIER_SKILLS for p in self.players])

    def get_mid_tier(self) -> int:
        return sum([p.skill in MID_TIER_SKILLS for p in self.players])

    def get_bottom_tier(self) -> int:
        return sum([p.skill in BOTTOM_TIER_SKILLS for p in self.players])

    def get_goalies(self) -> int:
        return sum([p.goalie_skill <= GOALIE_THRESHOLD for p in self.players])

    def __repr__(self):
        return (
            f"Team {self.name:11} "
            f"({self.practice_day} "
            f"{self.location} "
            f"size={len(self.players)}, "
            f"skill={self.get_skill():.2f}, "
            f"grade={self.get_grade():.2f})"
        )

    def __post_init__(self):
        validate_team_name(self)
        validate_team_day(self)
        validate_team_location(self)


@dataclass(frozen=True)
class Move:
    """A class representing moving a player from one team to another."""

    player: Player
    team_from: Optional[Team] = None
    team_to: Optional[Team] = None


@dataclass
class League:
    teams: List[Team]
    available_players: Set[Player]  # TODO: consider a PQ

    def __post_init__(self):
        self.scorer = CompositeScorer(self.players, self.teams)

    @property
    def players(self) -> List[Player]:
        """Return all players in the league."""
        return [p for team in self.teams for p in team.players] + list(
            self.available_players
        )

    def get_next_player(self) -> Player:
        """Gets the next highest-skilled available player for assignment."""
        return min(self.available_players, key=lambda p: p.skill)

    def add_player(self, player: Player, team: Team):
        # Note that the order matters. The scorer must run before the team changes.
        self.scorer.update_score_addition(player, team)
        team.players.add(player)

    def remove_player(self, player: Player, team: Team):
        # Note that the order matters. The scorer must run before the team changes.
        self.scorer.update_score_removal(player, team)
        team.players.remove(player)

    def apply_moves(self, moves: List[Move]) -> None:
        for move in moves:
            if move.team_from is not None:
                self.remove_player(move.player, move.team_from)
            else:
                self.available_players.remove(move.player)
            if move.team_to is not None:
                self.add_player(move.player, move.team_to)
            else:
                self.available_players.add(move.player)

    def undo_moves(self, moves: List[Move]) -> None:
        for move in moves:
            if move.team_to is not None:
                self.remove_player(move.player, move.team_to)
            else:
                self.available_players.remove(move.player)
            if move.team_from is not None:
                self.add_player(move.player, move.team_from)
            else:
                self.available_players.add(move.player)

    def reset_league(self) -> None:
        for team in self.teams:
            for player in list(team.players):
                self.apply_moves([Move(player=player, team_from=team, team_to=None)])

    def details(self) -> None:
        for key, scorer in self.scorer.scorers.items():
            print(f"{key:12}: {scorer.get_score():.3f}")

        team_info_dicts = []
        for team in self.teams:
            team_info_dict = {
                "name": team.name,
                "day": team.practice_day,
                "location": team.location,
                "size": len(team.players),
                "first_round": team.get_first_round(),
                "top_tier": team.get_top_tier(),
                "mid_tier": team.get_mid_tier(),
                "bottom_tier": team.get_bottom_tier(),
                "goalies": team.get_goalies(),
                "skill": team.get_skill(),
                "grade": team.get_grade(),
            }

            team_info_dicts.append(team_info_dict)
        team_info_df = pd.DataFrame.from_records(team_info_dicts)
        with pd.option_context("display.precision", 3):
            print(team_info_df.sort_values(by="name"))

    @classmethod
    def from_csvs(cls, player_csv: str, team_csv: str) -> "League":
        player_info = pd.read_csv(player_csv)
        team_info = pd.read_csv(team_csv)
        teams = {}
        for i, row in team_info.iterrows():
            teams[row["name"]] = Team(**row.to_dict())

        # Add all players as available to start
        players = []
        assigned_team_names = []
        for i, row in player_info.iterrows():
            player_dict = row.to_dict()
            team_name = player_dict.pop("team", None)
            player = Player.from_raw_dict(player_dict)
            assigned_team_names.append(team_name)
            players.append(player)

        league = cls(teams=list(teams.values()), available_players=set(players))

        # Now go and move them to their teams
        for player, team_name in zip(players, assigned_team_names):
            if not pd.isnull(team_name):
                if team_name not in teams:
                    if player.lock:
                        raise ValueError(
                            f"\033[1m"
                            f"Failed to add {player.first_name} {player.last_name} to Team "
                            f"{team_name}. This team was not found among the valid teams provided: "
                            f"{list(teams.keys())} and this player is locked. Please check your spelling, correct this in the "
                            "csv, or unlock this player and retry.\n"
                            f"\033[0m"
                        )
                    
                    print(
                        f"Failed to add {player.first_name} {player.last_name} to Team "
                        f"{team_name}. This team was not found in the team information: "
                        f"{list(teams.keys())}. This player is not locked, so will be assigned to a valid team"
                        "csv and retry.\n"
                    )
                    continue
                
                move = Move(player=player, team_from=None, team_to=teams[team_name])
                if breaks_practice_constraint(move):
                    if player.lock:
                        raise ValueError(
                            f"\033[1m"
                            f"Failed to add {player.first_name} {player.last_name} to Team "
                            f"{team_name}.\nThis player's practice info is incompatible with the team's "
                            f"practice info and the player is locked:\nPlayer's unavailable days: {player.unavailable_days}\n"
                            f"Player's disallowed locations: {player.disallowed_locations}\n"
                            f"Team info: {teams[team_name].name} practices at {teams[team_name].location} on {teams[team_name].practice_day}"
                            f"\nEither try unlocking the player or checking the practice "
                            "day and location before correcting this in the input csv and retrying. "
                            f"\033[0m"
                        )
                    print(
                        f"Failed to add {player.first_name} {player.last_name} to Team "
                        f"{team_name}.\nThis player's practice info is incompatible with the team's.\n"
                        f"This player is not locked, so will be assigned to a valid team\n"
                    )
                    continue
                league.apply_moves([move])

            else:
                if player.lock:
                    handle_error(
                        f"Player {player.first_name} {player.last_name} is locked but not assigned to a team. They will be assigned to a new team.\n"
                        , False 
                    )
                    

        return league

    def to_csv(self, player_csv: str) -> None:
        teams = []
        players = []
        for team in self.teams:
            team_dict = asdict(team)
            del team_dict["players"]
            teams.append(team_dict)
            print(team)
            for player in team.players:
                temp_player_dict = asdict(player)
                temp_player_dict["team"] = team.name
                # players.append(asdict(player) | {"team": team.name})
                players.append(temp_player_dict)

        for player in self.available_players:
            players.append(asdict(player))

        print(f"Saving player information to {player_csv}")
        pd.DataFrame.from_records(players).to_csv(player_csv, index=False)

    def __repr__(self):
        s = f"{len(self.teams)} Teams:\n"
        for team in self.teams:
            s += f"{team}\n"
        s += f"{len(self.available_players)} players available for assignment"
        return s



import argparse
import logging
import os

from geopy.extra.rate_limiter import RateLimiter
from geopy.geocoders import Nominatim
import numpy as np
import pandas as pd
from thefuzz import fuzz
from tqdm import tqdm
from cyslf.utils import handle_error

from cyslf.utils import DAY_MAP, FIELD_LOCATIONS, FIELD_MAP, get_dist
from cyslf.validation import request_validation, validate_file


geolocator = Nominatim(user_agent="cyslf")
geocode = RateLimiter(geolocator.geocode, min_delay_seconds=1 / 2)
tqdm.pandas()
pd.set_option("display.max_rows", None)

parser = argparse.ArgumentParser(
    description="Process raw registration forms into a standard csv."
)

parser.add_argument(
    "--division",
    "--div",
    "-d",
    type=str,
    help="division without season (e.g. 'Boys Grades 3-4'). Spelling needs to match form data.",
)

parser.add_argument(
    "--folder",
    "-f",
    type=str,
    help="Folder containing coach evaluations, parent requests, and registrations",
    default=None,
)

parser.add_argument(
    "--coach_evals",
    "-ce",
    type=str,
    default=None,
    help="Old registration csv. Where to pull coach evals + past teams from.",
)

parser.add_argument(
    "--parent_requests",
    "--par",
    type=str,
    default=None,
    help="Parent request form (practice preferences / teammate requests)",
)

parser.add_argument(
    "--registration", "--reg", type=str, help="Current registration csv.", default=None
)

parser.add_argument(
    "---matches",
    "-m",
    default=5,
    type=int,
    help="Number of potential name matches to display",
)

parser.add_argument(
    "--output_file",
    "-o",
    type=str,
    help="output csv file. eg 'b34-players.csv'",
)

parser.add_argument(
    "--replace",
    "-r",
    action="store_true",
    help="Overwrite the output file if it exists",
)


def _extract_num(column, dtype=float):
    """Extract numbers from a string column."""
    return column.str.extract("(\d+)", expand=False).astype(dtype)  # noqa: W605


def _extract_grade(column):
    """Extract the grade number from columns, setting P to -1 and K to 0"""
    grade = column.fillna("").astype(str).str[0]
    grade = grade.str.replace("P", "-1", regex=False)
    grade = grade.str.replace("K", "0", regex=False)
    return grade.astype(float)  # Allow missing grade for now


def _normalize_str(column):
    """Normalize string by removing non-alpha and lowercasing (partciularly for name matching)."""
    return column.str.lower().str.replace("[^a-zA-Z]", "", regex=True)


def _get_names(df):
    return ", ".join(df[["first_name", "last_name"]].agg(" ".join, axis=1))


# read coach eval file
def _load_existing_player_data(filename: str, division: str):
    print(f"\n===Reading existing player data from {filename}===")
    existing_players_raw = pd.read_csv(filename)
    print(f"{len(existing_players_raw)} players found in previous registration data.")
    # Only keep teams if the players were in the relevant division.
    # This is because a Spring 2nd-grader on Red shouldn't stay on Red, but
    # we do want to know their skill score.
    in_division = existing_players_raw["division"].str.contains(division)
    print(f"{in_division.sum()} were found in division: {division}")
    if in_division.sum() == 0:
        request_validation(
            "WARNING: Please confirm that the division is spelled correctly before proceeding"
        )
    existing_players_raw.loc[~in_division, "team"] = np.nan

    # changed to match coach evals
    column_map = {
        "rating_overall (1 is high)": "coach_skill",
        "rating_goalie": "goalie_skill",
        "lastname": "last_name",
        "firstname": "first_name",
        "team": "team",
    }
    for key, _ in column_map.items():
        if key not in existing_players_raw.columns:
            handle_error(f"Coach evals file is missing the \"{key}\" column.", True)
    existing_players = existing_players_raw.rename(columns=column_map)[
        column_map.values()
    ]

    # Players not already in the current division are coming from a younger division. Since they're
    # younger, we'll make their skill one point worse. In the future, maybe we can pick a more
    # principled offset
    print(
        "Lowering player skill by 1 for players that weren't previously in this division"
    )
    existing_players.loc[~in_division, "coach_skill"] += 1

    # Construct name for matching to current registration
    full_name = existing_players["first_name"] + existing_players["last_name"]
    existing_players["name_key"] = _normalize_str(full_name)

    # Handle duplicates by taking the highest score for a given name
    existing_players = (
        existing_players.drop(columns=["first_name", "last_name"])
        .sort_values(by="coach_skill")
        .drop_duplicates("name_key")
    )

    return existing_players

def _translate_practice_days(row, match1):
    ans = [day for day in DAY_MAP.values() if row[day] == match1]
    return "".join(ans)

def _translate_locations(row, match):
    # TODO: optimize
    ans = ""
    if row["east"] == match:
        for location in FIELD_MAP["East"]:
            ans += location + ", "
    if row["central"] == match:
        for location in FIELD_MAP["Central"]:
            ans += location + ", "
    if row["cambridgeport"] == match:
        for location in FIELD_MAP["Cambridgeport"]:
            ans += location + ", "
    if row["west"] == match:
        for location in FIELD_MAP["West"]:
            ans += location + ", "
    if row["north"] == match:
        for location in FIELD_MAP["North"]:
            ans += location + ", "
    if len(ans) > 0: ans = ans[:-2]
    return ans


def _load_parent_requests(filename):
    print(f"\n===Reading parent requests from {filename}===")
    parent_reqs = pd.read_csv(filename)
    print(f"{len(parent_reqs)} parent requests found")

    column_map = {
        "Player First Name (as it is written in your CYS/Sports Connect account)": "first_name",
        "Player Last Name": "last_name",
        "Preferred practice day(s) [Monday]": "M",
        "Preferred practice day(s) [Tuesday]": "T",
        "Preferred practice day(s) [Wednesday]": "W", 
        "Preferred practice day(s) [Thursday]": "R",
        "Preferred practice day(s) [Friday]": "F",
        "Practice Location(s) [East]": "east",
        "Practice Location(s) [Central]": "central",
        "Practice Location(s) [Cambridgeport]": "cambridgeport",
        "Practice Location(s) [West]": "west",
        "Practice Location(s) [North]": "north",
        "Teammate Request 1 First Name": "teammate_req1_firstname",
        "Teammate Request 1 Last Name": "teammate_req1_lastname",
        "Teammate Request 2 First Name": "teammate_req2_firstname",
        "Teammate Request 2 Last Name": "teammate_req2_lastname",
        "Teammate Request 3 First Name": "teammate_req3_firstname",
        "Teammate Request 3 Last Name": "teammate_req3_lastname",
        "I am willing to ": "extra_comment",
    }

    # Account for missing columns

    if "Player First Name (as it is written in your CYS/Sports Connect account)" not in parent_reqs.columns:
        handle_error("Missing first name column in parent requests file. Make sure the column name matches the docs.", True)
        
    if "Player Last Name" not in parent_reqs.columns:
        handle_error("Missing last name column in parent requests file. Make sure the column name matches the docs.", True)
        
    for key, val in column_map.items():
        if key not in parent_reqs.columns:
            print(f"Parent request file is missing column {key}. An empty column was added.")
            parent_reqs[val] = ''

    parent_reqs = parent_reqs.rename(columns=column_map)[column_map.values()]

    parent_reqs["name_key"] = _normalize_str(
        parent_reqs["first_name"] + parent_reqs["last_name"]
    )
    
    parent_reqs["preferred_days"] = parent_reqs.apply(lambda row: _translate_practice_days(row, "Ideal"), axis=1)
    
    parent_reqs["unavailable_days"] = parent_reqs.apply(lambda row: _translate_practice_days(row, "Impossible"), axis=1) 


    # concat all teammate request names for easy search 
    parent_reqs["teammate_requests"] = (parent_reqs["teammate_req1_firstname"].fillna('') + " " +  
                                        parent_reqs["teammate_req1_lastname"].fillna('') + ", " + 
                                        parent_reqs["teammate_req2_firstname"].fillna('') + " " + 
                                        parent_reqs["teammate_req2_lastname"].fillna('') + ", " + 
                                        parent_reqs["teammate_req3_firstname"].fillna('') + " " + 
                                        parent_reqs["teammate_req3_lastname"].fillna(''))

    parent_reqs["preferred_locations"] =  parent_reqs.apply(lambda row: _translate_locations(row, "Ideal"), axis=1)

    parent_reqs["disallowed_locations"]  = parent_reqs.apply(lambda row: _translate_locations(row, "Impossible"), axis=1) 


    parent_reqs = parent_reqs.drop(
        columns=["first_name", "last_name", "teammate_req1_firstname", "teammate_req1_lastname", "teammate_req2_firstname", "teammate_req2_lastname", "M", "T", "W", "R", "F"]
    )

    # Keep most recent entry
    return parent_reqs.drop_duplicates("name_key", keep="last")


# no change
def _load_registration_data(filename):
    print(f"\n===Reading registration data from {filename}===")
    registrations_raw = pd.read_csv(filename)

    if "latitude" not in registrations_raw.columns or "longitude" not in registrations_raw.columns:
        handle_error("Please use the convert addresses command before preparing player data. See documentation for more details.", True)
        
    
    # Merge the two school columns
    has_other_school = ~pd.isnull(registrations_raw["School Name other:"])
    registrations_raw.loc[has_other_school, "School Name"] = registrations_raw[
        has_other_school
    ]["School Name other:"]


    column_map = {
        "Player Last Name": "last_name",
        "Player First Name": "first_name",
        "Current Grade": "grade",
        "Parental assessment of player ability/athleticism:": "parent_skill",
        "School Name": "school",
        "longitude": "longitude",
        "latitude": "latitude",
    }
    for key, _ in column_map.items():
        if key not in registrations_raw.columns:
            handle_error(f"Registration file doesn't contain the \"{key}\" column.", True)
            
    
    registrations = registrations_raw.rename(columns=column_map)[column_map.values()]

    # Extract grade
    registrations["grade"] = _extract_grade(registrations["grade"])

    # Extract parent skill (they have opposite order from coach skill, so invert)
    registrations["parent_skill"] = _extract_num(registrations["parent_skill"])
    registrations["parent_skill"] = 11 - registrations["parent_skill"]

    # deal with kickstart players being rated 0
    registrations["parent_skill"] = registrations["parent_skill"].replace(11, 5)



    # Construct name for matching to current registration.
    full_name = registrations["first_name"] + registrations["last_name"]
    registrations["name_key"] = _normalize_str(full_name)

    print(f"{len(registrations)} registrations found")
    return registrations


def _merge_data(existing_players, parent_reqs, registrations):
    print("\n===Merging data===")

    ordered_columns = [
        "id",
        "last_name",
        "first_name",
        "grade",
        "team",
        "coach_skill",
        "parent_skill",
        "goalie_skill",
        "preferred_days",
        "unavailable_days",
        "preferred_locations",
        "disallowed_locations",
        "backup_locations",
        "teammate_requests",
        "lock",
        "emailed_parents",
        "school",
        "comment",
    ]
    players = registrations.merge(existing_players, how="left", on="name_key").merge(
        parent_reqs, how="left", on="name_key"
    )
    players["id"] = players.index.values

    # Lock players to a team if they already have one
    players["lock"] = False
    # missing_team = pd.isnull(players.team)
    # players.loc[missing_team, "lock"] = False
    # print(f"{(~missing_team).sum()} players locked onto their teams from last season")

    players["emailed_parents"] = False

    # Set backup field preferences based on location
    # If parents didn't fill out the request form with a location preference, or if the preferred
    # locations don't provide enough options, we'd like to have a fallback based on home address.
    # Using geopy, we figure out the 3 closest fields not already covered by the
    # preferred / disallowed locations and stick those in a column.
    backup_locations = []
    for _, row in players.iterrows():
        lat, long = row[["latitude", "longitude"]]
        if pd.isnull(lat) or pd.isnull(long):
            backup_locations.append("")
            continue
        field_df = pd.DataFrame.from_dict(
            FIELD_LOCATIONS, orient="index", columns=["lat", "long"]
        )
        # Don't think about fields if they're already in preferred/disallowed locations
        mask = field_df.index.map(
            lambda f: f in row.fillna("")["preferred_locations"]
            or f in row.fillna("")["disallowed_locations"]
        ).values.astype(bool)
        field_df = field_df[~mask]
        if len(field_df) <= 3:
            backup_locations.append("")
            continue
        field_df["distance"] = field_df.apply(
            lambda r: get_dist(lat, long, r["lat"], r["long"]), axis=1
        )
        field_df = field_df.sort_values(by="distance")
        backup_locations.append(", ".join(field_df.head(3).index.values))
    players["backup_locations"] = backup_locations

    players["comment"] = (
        # players["comment"].fillna("") + " || " + players["extra_comment"].fillna("")
        players["extra_comment"].fillna("")
    )

    # Fill missing columns with nans
    for col in ordered_columns:
        if col not in players.columns:
            players[col] = np.nan

    players = players[ordered_columns]

    print(f"Finished merging. Total # of players: {len(players)}")
    return players


def _final_sweep(players):
    print("\n===Running final checks===")

    normalized_name = _normalize_str(players["first_name"] + players["last_name"])
    name_counts = normalized_name.value_counts()
    duplicate_names = name_counts[name_counts > 1].index.values
    has_duplicate = normalized_name.isin(duplicate_names)
    if has_duplicate.sum() > 0:
        print(
            f"{has_duplicate.sum()} rows have names that match other rows, so please check if "
            "they're duplicates:"
        )
        print(
            players[has_duplicate].sort_values(by="last_name")[
                ["id", "first_name", "last_name"]
            ]
        )

    missing_last_name = pd.isnull(players.last_name)
    if missing_last_name.sum() > 0:
        print(f"{missing_last_name.sum()} players are missing a last name:")
        print(players[missing_last_name][["id", "first_name", "last_name"]])

    missing_first_name = pd.isnull(players.first_name)
    if missing_first_name.sum() > 0:
        print(f"{missing_first_name.sum()} players are missing a first name:")
        print(players[missing_first_name][["id", "first_name", "last_name"]])

    missing_grade = pd.isnull(players.grade)
    if missing_grade.sum() > 0:
        print(
            f"{missing_grade.sum()} players are missing a grade: "
            f"{_get_names(players[missing_grade])}"
        )

    missing_skill = pd.isnull(players.coach_skill) & pd.isnull(players.parent_skill)
    if missing_skill.sum() > 0:
        print(
            f"{missing_skill.sum()} players don't have a coach or parent skill. Automatically "
            "assigning them a skill of 5"
        )
        players.loc[missing_skill, "parent_skill"] = 5

    missing_goalie_skill = pd.isnull(players.goalie_skill)
    if missing_goalie_skill.sum() > 0:
        print(
            f"{missing_goalie_skill.sum()} players don't have a goalie skill. Automatically "
            "assigning them a skill of 6 (does not play goalie)."
        )
        players.loc[missing_goalie_skill, "goalie_skill"] = 6


def cross_match_names(df1, df2, suffixes):
    # dfs are assumed to have name_key columns

    matched_names = np.intersect1d(df1["name_key"].values, df2["name_key"].values)
    unmatched_df1 = df1[~df1["name_key"].isin(matched_names)]
    unmatched_df2 = df2[~df2["name_key"].isin(matched_names)]
    cross = unmatched_df1.merge(unmatched_df2, how="cross", suffixes=suffixes)
    keys = [f"name_key{s}" for s in suffixes]
    if cross.shape[0] == 0: return cross
    cross["score"] = cross.apply(lambda r: fuzz.ratio(r[keys[0]], r[keys[1]]), axis=1)
    return cross.sort_values(by="score", ascending=False)[[keys[0], keys[1], "score"]]


def main():
    args = parser.parse_args()
    
    #TODO: make argument validating logic simpler/more seamless

    if not args.division:
        handle_error("No division provided.", True)
    
    if not args.registration: 
        if args.folder: 
            args.registration = args.folder + "/registration.csv"
        else: 
            handle_error("Please provide a registration file.", True)

    if not args.coach_evals: 
        if args.folder:
            args.coach_evals = args.folder + "/coach_evals.csv"
        else:
            raise handle_error("Please provide a coach evaluations file.", True)
        
    if not args.output_file: 
        if args.folder:
            args.output_file = args.folder + "/prepared_player_data.csv"
        else:
            handle_error("Please provide a valid output file path.", True)
        
    if not args.parent_requests and args.folder: 
        args.parent_requests = args.folder + "/parent_requests.csv" 
    
    try: 
        validate_file(args.registration)
    except:
        handle_error(f"Invalid registration file path {args.registration}.", True)
    try:   
        validate_file(args.coach_evals)
    except:
        handle_error(f"Invalid coach evaluations file path {args.coach_evals}.", True)

    try: 
        validate_file(args.parent_requests)
    except:
        args.parent_requests = None

    # Load current registrations
    registrations = _load_registration_data(args.registration)

    # Pull team and coach score from spring 2022
    existing_players = _load_existing_player_data(args.coach_evals, args.division)
    existing_player_match = existing_players["name_key"].isin(registrations["name_key"])
    lower_division_match = pd.isnull(existing_players["team"]) & existing_player_match
    print(
        f"{existing_player_match.sum()} / {len(registrations)} registrations were matched to "
        f"existing player data. {lower_division_match.sum()} came from a lower division."
    )
    cross = cross_match_names(registrations, existing_players, ("_reg", "_old"))
    print(
        "If names are spelled differently across forms, they can't be automatically matched. \n"
        f"Please check the following list of potential name matches (printing top {args.matches}):"
    )
    print(cross.head(args.matches))
    print("Please edit the forms so the names match or manually update the player csv")

    # Load parent requests
    if args.parent_requests is not None:
        print("Parent requests found")
        parent_reqs = _load_parent_requests(args.parent_requests)
        parent_req_match = registrations["name_key"].isin(parent_reqs["name_key"])
        print(
            f"{parent_req_match.sum()} / {len(registrations)} players had matching parent requests"
        )
        cross = cross_match_names(registrations, parent_reqs, ("_reg", "_parent"))
        print(
            "If names are spelled differently across forms, they can't be automatically matched. \n"
            "Please check the following list of potential name matches (printing top "
            f"{args.matches}):"
        )
        print(cross.head(args.matches))
        print(
            "Please edit the forms so the names match or manually update the player csv"
        )
    else:
        print("No parent requests provided.")
        parent_reqs = pd.DataFrame(
            columns=[
                "name_key",
                "disallowed_locations",
                "preferred_locations",
                "preferred_days",
                "unavailable_days",
                "extra_comment",
                "teammate_requests",
            ]
        )

    # Merge the old data into the current registration data.
    players = _merge_data(existing_players, parent_reqs, registrations)

    # Do a final cleaning pass
    _final_sweep(players)

    if not args.output_file and args.folder: args.output_file = args.folder + "/prepared_player_data.csv"

    if args.replace or not os.path.exists(args.output_file):
        print(f"Saving to {args.output_file}")
        players.to_csv(args.output_file, index=False)
    else:
        handle_error(f"Not saving. {args.output_file} already exists. Use -r to replace", True)


if __name__ == "__main__":
    main()



import argparse
import logging
import os

from geopy.extra.rate_limiter import RateLimiter
from geopy.geocoders import Nominatim
import numpy as np
import pandas as pd
from thefuzz import fuzz
from tqdm import tqdm
from .utils import handle_error

from cyslf.utils import DAY_MAP, FIELD_LOCATIONS, FIELD_MAP, get_dist
from cyslf.validation import request_validation, validate_file


geolocator = Nominatim(user_agent="cyslf")
geocode = RateLimiter(geolocator.geocode, min_delay_seconds=1 / 2)
tqdm.pandas()
pd.set_option("display.max_rows", None)

parser = argparse.ArgumentParser(
    description="Process raw registration forms into a standard csv."
)

# keep
parser.add_argument(
    "--registration", "--reg", type=str, help="Current registration csv."
)

parser.add_argument(
    "--folder",
    "-f",
    type=str,
    help="Folder containing registrations",
    default=None,
)

def _lookup_location(address):
    # If an address is poorly formed, geopy gives sad-looking warnings,
    # so let's temporarily disable this.
    logging.getLogger("geopy").setLevel(logging.ERROR)
    location = geocode(address)
    logging.getLogger("geopy").setLevel(logging.WARNING)
    if location:
        return location.latitude, location.longitude
    else:
        print("\033[1m" + f"Failed to find address: {address}" + "\033[0m")
        return np.nan, np.nan

def _convert_addresses(filename):
    print(f"\n===Reading registration data from {filename}===")
    registrations_raw = pd.read_csv(filename)
    registrations_raw["Postal Code"] = registrations_raw["Postal Code"].apply(lambda x: "0" + str(x))

    # Look up player latitude / longitude
    print("Converting addresses to latitude / longitude")
    registrations_raw["Postal Code"] = registrations_raw["Postal Code"].astype(str)
    registrations_raw["Address"] = registrations_raw[
        ["Street", "City", "Region", "Postal Code"]
    ].agg(", ".join, axis=1)

    

    registrations_raw["Location"] = registrations_raw["Address"].progress_apply(
        _lookup_location
    )
    registrations_raw["latitude"] = registrations_raw["Location"].apply(lambda x: x[0])
    registrations_raw["longitude"] = registrations_raw["Location"].apply(lambda x: x[1])

    registrations_raw = registrations_raw.drop(
        columns=["Location", "Address"]
    )

    registrations_raw.to_csv(filename, index = False)

    return 

def main():
    args = parser.parse_args()
    if not args.registration: 
        if args.folder:
            args.registration = args.folder + "/registration.csv"
        else: 
            handle_error("Please provide a registration file.", True)
            
    try: 
        validate_file(args.registration)
    except:
        handle_error(f"Invalid registration file path {args.registration}.", True)

    _convert_addresses(args.registration)



if __name__ == "__main__":
    main()



import itertools as it
from typing import Dict, List, Optional

from .constraints import breaks_constraints
from .models import League, Move, Player
from .utils import handle_error


def find_best_moves(
    player: Player,
    league: League,
    depth: int = 1,
    weights: Optional[Dict[str, float]] = None,
) -> List[Move]:
    queue = []
    best_moves = []
    best_score: float = -1
    old_team = None 

    # removed this because jason wants locked players who are on incompatible teams to be re-assigned
    # if player.lock: return []
    
    for team in league.teams:
        queue.append([Move(player=player, team_from=old_team, team_to=team)])


    while len(queue) > 0:
        proposed_moves = queue.pop()
        if breaks_constraints(proposed_moves):            
            continue
        league.apply_moves(proposed_moves)
        score = league.scorer.get_score(weights=weights)
        league.undo_moves(proposed_moves)
        if score > best_score:
            best_score = score
            best_moves = proposed_moves

        if len(proposed_moves) < depth:
            last_team = proposed_moves[-1].team_to
            if last_team is None:
                handle_error(
                    f"Proposed move tried to unassign a player, which is not allowed. Please check "
                    f"move proposal creation. Move: {proposed_moves[-1]}", True
                )

            for p, t in it.product(last_team.players, league.teams):
                if p.lock: continue
                moved_players = [move.player for move in proposed_moves]
                if p in moved_players or t.name == last_team.name:
                    continue
                queue.append(
                    [*proposed_moves, Move(player=p, team_from=last_team, team_to=t)]
                )
    if len(best_moves) == 0:
        handle_error(
            f"Failed to place Player {player.first_name} {player.last_name} on a team. This is "
            "likely due to unsatisfiable scheduling constraints (There is no valid team for this player). Please try manually assigning "
            "this player before retrying.", True
        )
    return best_moves



"""
Scorers

Scorers should score a given league setup on a scale from 0 to 1.

The main goal here is that "better league arrangement" gets a higher score. Each class implicitly
exposes an interface containing:
* update_score_addition(player, team): update internal metrics associated with adding player->team
* update_score_removal(player, team): update internal metrics associated with removing player->team
* get_score(): return a score between 0 and 1

base.py contains some abstract base scorers that we use here.

CompositeScorer is the primary exported scorer -- it goes scores everything in SCORER_MAP
"""


from typing import TYPE_CHECKING, Dict, List, Optional

from ..utils import (
    BOTTOM_TIER_SKILLS,
    FIRST_ROUND_SKILL,
    GOALIE_THRESHOLD,
    MID_TIER_SKILLS,
    TOP_TIER_SKILLS,
)
from .base import CountParityScorer, CountScorer, MeanParityScorer


if TYPE_CHECKING:
    from .models import Player, Team


# CONVENIENCE SCORERS
class PracticeDayScorer(CountScorer):
    def _count_player(self, player: "Player", team: "Team") -> bool:
        return team.practice_day in player.preferred_days


class LocationScorer:
    def __init__(self, players: List["Player"], teams: List["Team"]):
        self.league_size = len(players)
        self.count = 0.0
        for t in teams:
            self.count += sum([self._count_player(p, t) for p in t.players])

    def _count_player(self, player: "Player", team: "Team") -> float:
        if team.location in player.preferred_locations:
            return 1
        elif team.location in player.backup_locations:
            return 0.5
        else:
            return 0

    def update_score_addition(self, player: "Player", team: "Team"):
        self.count += self._count_player(player, team)

    def update_score_removal(self, player: "Player", team: "Team"):
        self.count -= self._count_player(player, team)

    def get_score(self) -> float:
        return self.count / self.league_size


class TeammateScorer:
    """Scorer for up to 1 teammate request

    We check friend requests by simply checking if {first_name} {last_name} is in the request
    string. We also award at most one friend request per player. This means when adding / removing
    players, we only want to increment / decrement if that player was uniquely requested by someone
    else on the team.
    """

    def __init__(self, players: List["Player"], teams: List["Team"]):
        # Assumes teams are empty, which is technically how we do it
        self.league_size = len(players)
        self.friend_matches = 0

    def update_score_addition(self, player: "Player", team: "Team"):
        request_satisfied = False
        for p in team.players:
            name = f"{p.first_name} {p.last_name}"
            if name in player.teammate_requests and not request_satisfied:
                self.friend_matches += 1
                request_satisfied = True
            if f"{player.first_name} {player.last_name}" in p.teammate_requests:
                player_is_unique = True
                for q in team.players:
                    if p == q:
                        continue
                    if f"{q.first_name} {q.last_name}" in p.teammate_requests:
                        player_is_unique = False
                        break
                if player_is_unique:
                    self.friend_matches += 1

    def update_score_removal(self, player: "Player", team: "Team"):
        request_satisfied = False
        for p in team.players:
            name = f"{p.first_name} {p.last_name}"
            if name in player.teammate_requests and not request_satisfied:
                self.friend_matches -= 1
                request_satisfied = True
            if f"{player.first_name} {player.last_name}" in p.teammate_requests:
                player_is_unique = True
                for q in team.players:
                    if p == q or player == q:
                        continue
                    if f"{q.first_name} {q.last_name}" in p.teammate_requests:
                        player_is_unique = False
                        break
                if player_is_unique:
                    self.friend_matches -= 1

    def get_score(self) -> float:
        return self.friend_matches / self.league_size


# PARITY SCORERS
class SizeScorer(CountParityScorer):
    def _count_player(self, player: "Player") -> bool:
        return True


class FirstRoundScorer(CountParityScorer):
    def _count_player(self, player: "Player") -> bool:
        return player.skill == FIRST_ROUND_SKILL


class TopTierScorer(CountParityScorer):
    def _count_player(self, player: "Player") -> bool:
        return player.skill in TOP_TIER_SKILLS


class MidTierScorer(CountParityScorer):
    def _count_player(self, player: "Player") -> bool:
        return player.skill in MID_TIER_SKILLS


class BottomTierScorer(CountParityScorer):
    def _count_player(self, player: "Player") -> bool:
        return player.skill in BOTTOM_TIER_SKILLS


class GoalieScorer(CountParityScorer):
    def _count_player(self, player: "Player") -> bool:
        return player.goalie_skill <= GOALIE_THRESHOLD


class GradeScorer(MeanParityScorer):
    def _get_value(self, player: "Player") -> bool:
        return player.grade


class SkillScorer(MeanParityScorer):
    def _get_value(self, player: "Player") -> bool:
        return player.skill


SCORER_MAP = {
    "skill": SkillScorer,
    "grade": GradeScorer,
    "size": SizeScorer,
    "location": LocationScorer,
    "practice_day": PracticeDayScorer,
    "teammate": TeammateScorer,
    "first_round": FirstRoundScorer,
    "top": TopTierScorer,
    "mid": MidTierScorer,
    "bottom": BottomTierScorer,
    "goalie": GoalieScorer,
}

DEFAULT_WEIGHTS = {
    "skill": 0.3,
    "grade": 0.3,
    "size": 0.15,
    "location": 0.05,
    "practice_day": 0.05,
    "first_round": 0.15,
    "top": 0.1,
    "mid": 0.1,
    "bottom": 0.1,
    "teammate": 0.1,
    "goalie": 0.1,
}

# COMPOSITE SCORER


class CompositeScorer:
    def __init__(self, players: List["Player"], teams: List["Team"]):
        self.scorers = {}
        for key, scorer in SCORER_MAP.items():
            self.scorers[key] = scorer(players, teams)

    def update_score_addition(self, player: "Player", team: "Team"):
        for scorer in self.scorers.values():
            scorer.update_score_addition(player, team)  # type: ignore

    def update_score_removal(self, player: "Player", team: "Team"):
        for scorer in self.scorers.values():
            scorer.update_score_removal(player, team)  # type: ignore

    def get_score(self, weights: Optional[Dict[str, float]] = None) -> float:
        if weights is None:
            weights = DEFAULT_WEIGHTS
        score: float = 0
        total_weight: float = 0
        s = ""
        verbose = False
        for scorer_key, weight in weights.items():
            scorer = self.scorers[scorer_key]
            score += weight * scorer.get_score()  # type: ignore
            total_weight += weight
            if verbose:
                s += f"{scorer.get_score():.3f} "  # type: ignore
        if verbose:
            print(s, list(weights.keys()))
        return score / total_weight



from typing import TYPE_CHECKING, List


if TYPE_CHECKING:
    from .models import Move


def breaks_constraints(moves: List["Move"]) -> bool:
    for move in moves:
        if breaks_practice_constraint(move):
            return True
        # if move.player.lock:
        #     return True
    return False


def breaks_practice_constraint(move: "Move") -> bool:
    return (
        move.team_to is None
        or move.team_to.practice_day in move.player.unavailable_days
        or move.team_to.location in move.player.disallowed_locations
    )



import argparse
import configparser
import os
import sys

from tqdm import tqdm

from cyslf.models import League
from cyslf.optimize import find_best_moves
from cyslf.scorers import DEFAULT_WEIGHTS
from cyslf.validation import validate_file
from cyslf.utils import handle_error


parser = argparse.ArgumentParser(description="Assign players to teams")
parser.add_argument(
    "--input_player_csv",
    "-i",
    type=str,
    help="input player csv. eg 'boys34-players.csv'",
)

parser.add_argument(
    "--output_player_csv",
    "-o",
    type=str,
    help="output player csv. eg 'boys34-results.csv'",
)

parser.add_argument(
    "--team_csv",
    "-t",
    type=str,
    help="team csv. should contain team information (name, location, etc)",
)

parser.add_argument(
    "--folder",
    "-f",
    type=str,
    help="Folder containing coach evaluations, parent requests, and registrations",
    default=None,
)

parser.add_argument("--config", "-c", type=str, help="config file with scorer weights.")
parser.add_argument("--depth", "-d", type=int, default=2, help="search depth")
parser.add_argument(
    "--replace",
    "-r",
    action="store_true",
    help="Overwrite the output file if it exists",
)


def main():
    args = parser.parse_args()
    if not args.input_player_csv:
        if args.folder:
            args.input_player_csv = args.folder + "/prepared_player_data.csv"
        else:
            handle_error("No prepared player data file or folder containing it provided.", True)

    if not args.output_player_csv:
        if args.folder:
            args.output_player_csv = args.folder + "/final_teams.csv"
        else:
            handle_error("No output file or folder provided.", True)
    if not args.team_csv:
        if args.folder:
            args.team_csv = args.folder + "/teams.csv"
        else:
            handle_error("No teams file or folder provided", True)
    if not args.config:
        if args.folder:
            try:
                validate_file(args.folder + "/weights.txt")
                args.config = args.folder + "/weights.txt"
            except: pass

    try: validate_file(args.input_player_csv)
    except:
        handle_error(f"Invalid prepared player data file path {args.input_player_csv}.", True)
    try: validate_file(args.team_csv)
    except:
        raise handle_error(f"Invalid teams file path {args.teams_csv}.", True)

    if not args.replace and os.path.exists(args.output_player_csv):
        handle_error(
            f"{args.output_player_csv} already exists. Use `-r` to replace it", True
        )

    print(
        f"Loading league from {args.input_player_csv} (players) and {args.team_csv} (teams)\n"
    )
    league = League.from_csvs(args.input_player_csv, args.team_csv)

    weights = None
    if args.config is not None:
        try: validate_file(args.config)
        except: handle_error(f"Invalid weights file path {args.config}.", True)
        config = configparser.ConfigParser(inline_comment_prefixes="#")
        config.read(args.config)
        weights = {}
        for key in config["weights"].keys():
            weights[key] = float(config["weights"][key])
    else:
        weights = DEFAULT_WEIGHTS
    print(f"Using weights {weights}")
    print(f"Using depth {args.depth}")
    print(f"Starting to assign {len(league.available_players)} available players.")
    for _ in tqdm(range(len(league.available_players))):
        player = league.get_next_player()
        best_moves = find_best_moves(player, league, depth=args.depth, weights=weights)
        league.apply_moves(best_moves)
    league.details()

    league.to_csv(args.output_player_csv)


if __name__ == "__main__":
    main()
